import sys
import threading


class ProgressBar:
    def __init__(
        self,
        total: int,
        program_name: str,
        bar_length=20,
        done_message="Finished!",
        update_interval=1,
        cost=None,
    ):
        self.total = total
        self.bar_length = bar_length
        self.done_message = done_message
        self.update_interval = update_interval
        self.current = 0
        self.lock = threading.Lock()

        print(f"\n\x1b[32m----Starting {program_name}----\x1b[0m\n")

        self.update(0, cost)

    def update(self, current: int, cost):
        with self.lock:
            self.current = current
            fraction = self.current / self.total

            arrow = int(fraction * self.bar_length) * "-" + ">"
            padding = (self.bar_length - len(arrow)) * " "

            if self.current == self.total:
                ending = f"\n\n{self.done_message}\x1b[0m\n\n"
            else:
                ending = "\r"

            completed = "\x1b[32m" if self.current == self.total else "\x1b[0m"

            print(
                f"{completed}Progress: [{arrow}{padding}] {fraction*100:.2f}% Cost: {cost}",
                end=ending,
            )
            sys.stdout.flush()

    def increment(self, step=1, cost=None):
        self.update(self.current + step, cost)



from typing import Callable, List, Tuple

import numpy as np
import pandas as pd


def balance_classes(df, y_col):
    # Group the data by the class column
    grouped = df.groupby(y_col)

    # Find the size of the smallest class
    min_size = grouped.size().min()

    # Resample all classes to match the size of the smallest class
    balanced_groups = [group.sample(min_size, random_state=42) for _, group in grouped]

    # Concatenate the resampled classes back together
    balanced_data = pd.concat(balanced_groups)

    return balanced_data


def generateDataset(
    name: str,
    num_input_variables: int,
    x_funcs: List[Callable[..., np.ndarray]],
    x_func_params: List[Tuple],
    y_func: Callable[[List[np.ndarray]], np.ndarray],
    balance_dataset: bool = True,
) -> None:
    def generate_data():
        input_variables: List[np.ndarray] = []
        for i in range(num_input_variables):
            input_col = x_funcs[i](*x_func_params[i])
            input_variables.append(input_col)

        y_col = y_func(input_variables)

        temp = {f"input_{i}": input_variables[i] for i in range(len(input_variables))}
        temp["y_col"] = y_col

        return pd.DataFrame.from_dict(temp)

    df = generate_data()

    # Determine the required size
    required_size = x_func_params[0][2]

    if balance_dataset:
        balanced_df = balance_classes(df, "y_col")

        # Keep generating data and balancing until we have enough samples
        while len(balanced_df) < required_size:
            new_df = generate_data()
            balanced_new_df = balance_classes(new_df, "y_col")
            balanced_df = pd.concat([balanced_df, balanced_new_df])

        # Shuffle and sample to get the exact required size
        balanced_df = balanced_df.sample(n=required_size, random_state=42).reset_index(
            drop=True
        )
        balanced_df.to_csv(name, index=False)
    else:
        df = df.sample(n=required_size, random_state=42).reset_index(drop=True)
        df.to_csv(name, index=False)


# Example usage (assuming x_funcs, x_func_params, and y_func are defined)
# generateDataset('output.csv', num_input_variables, x_funcs, x_func_params, y_func)



import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Tuple

import numpy as np
import pandas as pd

from .progress_bar import ProgressBar


def process_chunk(
    df_chunk: pd.DataFrame,
    cols: pd.Index,
    p: ProgressBar,
    lock: threading.Lock,
    num_out_classes: int,
) -> np.ndarray:
    data_points = []
    for _, row in df_chunk.iterrows():
        inputs = [row[cols[j]] for j in range(len(cols) - 1)]
        output = row[cols[-1]]
        expected_outputs = [0.0] * num_out_classes
        expected_outputs[int(output)] = 1.0
        data_points.append(
            (
                np.array(inputs, dtype=np.float64),
                np.array(expected_outputs, dtype=np.float64),
            )
        )
        with lock:
            p.increment()

    # Convert list of tuples to a structured numpy array
    data_points = np.array(data_points, dtype=object)
    return data_points


def trainTestSplit(
    df: pd.DataFrame, num_out_classes: int, ratio: float = 0.8, num_threads: int = 4
) -> Tuple[np.ndarray, np.ndarray]:
    data_points = []
    cols = df.columns
    p = ProgressBar(total=len(df), program_name="PAIN")
    lock = threading.Lock()

    chunk_size = len(df) // num_threads
    chunks = [df.iloc[i : i + chunk_size] for i in range(0, len(df), chunk_size)]

    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        future_to_chunk = {
            executor.submit(process_chunk, chunk, cols, p, lock, num_out_classes): chunk
            for chunk in chunks
        }

        for future in as_completed(future_to_chunk):
            data_points.extend(future.result())

    train_data = data_points[: int(ratio * len(data_points))]
    test_data = data_points[int(ratio * len(data_points)) :]

    return np.array(train_data), np.array(test_data)



from .LinearRegression.LinearRegression import LinearRegression
from .LinearRegression.MultiVariateLinearRegression import \
    MultivariateLinearRegression
from .PolynomialRegression.PolynomialRegression import PolynomialRegression



from setuptools import find_packages, setup

setup(
    name="SCTW_DS",
    version="0.3.2",
    packages=find_packages(),  # Automatically includes all sub-packages
    install_requires=[
        "numpy>=1.21.0",  # For numerical operations
        "pandas>=1.3.0",  # For data manipulation and analysis
    ],
    entry_points={
        "console_scripts": [
            # Example: 'my_script=my_package.module:main_func'
            # Add any CLI commands if you have any scripts to expose
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
)



import os
import pickle
from concurrent.futures import ProcessPoolExecutor, as_completed
from multiprocessing import set_start_method, shared_memory
from typing import List

import numpy as np

from ..Utils.progress_bar import ProgressBar
from .Layer import Layer

global_layers: List[Layer] = []
global_weights = []
global_biases = []

set_start_method("fork")


def initialize_shared_memory(layers):
    global global_layers, global_weights, global_biases
    global_layers = layers
    global_weights = []
    global_biases = []

    for layer in layers:
        weight_shm = shared_memory.SharedMemory(create=True, size=layer.weights.nbytes)
        bias_shm = shared_memory.SharedMemory(create=True, size=layer.biases.nbytes)
        global_weights.append(weight_shm)
        global_biases.append(bias_shm)
        np.copyto(
            np.ndarray(
                layer.weights.shape, dtype=layer.weights.dtype, buffer=weight_shm.buf
            ),
            layer.weights,
        )
        np.copyto(
            np.ndarray(
                layer.biases.shape, dtype=layer.biases.dtype, buffer=bias_shm.buf
            ),
            layer.biases,
        )


def cleanup_shared_memory():
    global global_weights, global_biases, global_layers
    for shm in global_weights:
        shm.close()
        shm.unlink()
    for shm in global_biases:
        shm.close()
        shm.unlink()

    global_weights = []
    global_biases = []
    global_layers = []


def _update_batch(batch: np.ndarray):
    global global_layers, global_weights, global_biases

    nabla_w = [np.zeros(layer.weights.shape) for layer in global_layers]
    nabla_b = [np.zeros(layer.biases.shape) for layer in global_layers]

    for data_point in batch:
        # Forward pass
        activation = data_point[0]
        activations = [activation]
        zs = []

        for i, layer in enumerate(global_layers):
            weights = np.ndarray(
                layer.weights.shape,
                dtype=layer.weights.dtype,
                buffer=global_weights[i].buf,
            )
            biases = np.ndarray(
                layer.biases.shape,
                dtype=layer.biases.dtype,
                buffer=global_biases[i].buf,
            )

            z = np.dot(activation, weights) + biases
            zs.append(z)
            activation = layer._activationFunction(z)
            activations.append(activation)

        # Backward pass
        last_layer = global_layers[-1]
        delta = last_layer.nodeCostDerivative(
            activations[-1], np.array(data_point[1])
        ) * global_layers[-1]._activationFunctionDerivative(zs[-1])
        nabla_b[-1] += delta.astype(float)
        nabla_w[-1] += np.outer(activations[-2], delta).astype(float)

        for layer_idx in range(2, len(global_layers) + 1):
            z = zs[-layer_idx]
            sp = global_layers[-layer_idx]._activationFunctionDerivative(z)
            weights_next = np.ndarray(
                global_layers[-layer_idx + 1].weights.shape,
                dtype=global_layers[-layer_idx + 1].weights.dtype,
                buffer=global_weights[-layer_idx + 1].buf,
            )
            delta = np.dot(delta, weights_next.T) * sp
            nabla_b[-layer_idx] += delta
            nabla_w[-layer_idx] += np.outer(activations[-layer_idx - 1], delta)

    return nabla_w, nabla_b


class NeuralNetwork:
    def __init__(self, layers: List[Layer]) -> None:
        self.layers = layers
        self.velocity_w = [np.zeros_like(layer.weights) for layer in layers]
        self.velocity_b = [np.zeros_like(layer.biases) for layer in layers]

        initialize_shared_memory(layers)

    def calculateOutputs(self, inputs: np.ndarray) -> np.ndarray:
        for layer in self.layers:
            inputs = layer.calculateOutputs(inputs)
        return inputs

    def _classify(self, inputs: np.ndarray) -> np.intp:
        outputs = self.calculateOutputs(inputs)
        return np.argmax(outputs)

    def classify(self, inputs: np.ndarray) -> List[np.intp]:
        preds: List[np.intp] = []
        for point in inputs:
            pred = self._classify(point[0])
            preds.append(pred)
        return preds

    def _cost(self, data_point: np.ndarray) -> np.float64:
        y_pred = self.calculateOutputs(np.array(data_point[0]))
        y_true = data_point[1]
        output_layer = self.layers[-1]
        cost = 0.0
        cost += output_layer.nodeCost(y_pred, y_true)
        return cost

    def cost(self, data_points: np.ndarray) -> np.float64:
        total_cost = np.float64(0.0)
        for data_point in data_points:
            total_cost += self._cost(data_point)
        return total_cost / len(data_points)

    def learn(
        self,
        training_data: np.ndarray,
        learning_rate: float,
        batch_size: int,
        momentum: float,
    ) -> np.float64:
        batches = [
            training_data[k : k + batch_size]
            for k in range(0, len(training_data), batch_size)
        ]

        initialize_shared_memory(self.layers)
        with ProcessPoolExecutor(max_workers=os.cpu_count()) as executor:
            future_to_batch = {
                executor.submit(_update_batch, batch): batch for batch in batches
            }

            nabla_w_total = [np.zeros_like(layer.weights) for layer in self.layers]
            nabla_b_total = [np.zeros_like(layer.biases) for layer in self.layers]

            for future in as_completed(future_to_batch):
                nabla_w, nabla_b = future.result()
                for i in range(len(self.layers)):
                    nabla_w_total[i] += nabla_w[i]
                    nabla_b_total[i] += nabla_b[i]

        global global_layers
        for i, layer in enumerate(self.layers):
            self.velocity_w[i] = (
                momentum * self.velocity_w[i]
                - (learning_rate / batch_size) * nabla_w_total[i]
            )
            self.velocity_b[i] = (
                momentum * self.velocity_b[i]
                - (learning_rate / batch_size) * nabla_b_total[i]
            )
            layer.weights += self.velocity_w[i]
            layer.biases += self.velocity_b[i]

        cleanup_shared_memory()

        return self.cost(training_data)

    def train(
        self,
        iterations: int,
        data_points: np.ndarray,
        learning_rate: float,
        batch_size: int,
        momentum: float = 0.9,
        # decay: float = 1e-7,
    ) -> List[np.float64]:
        costs: List[np.float64] = []
        progress = ProgressBar(total=iterations, program_name="training.")
        for _ in range(iterations):
            cost = self.learn(data_points, learning_rate, batch_size, momentum)
            progress.increment(cost=cost)
            costs.append(cost)

        return costs

    def save_model(self, file_path: str):
        with open(file_path, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def load_model(cls, file_path: str) -> "NeuralNetwork":
        with open(file_path, "rb") as file:
            return pickle.load(file)
